// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.

// Package version keeps the version number of the client package.
package version

// Version is the version of this client package that is communicated to the server.
const Version = "1.2.0"
