// Copyright 2023 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

package artifact

import (
	"io"

	gzip "github.com/klauspost/pgzip"
)

type CompressorGzip struct {
}

func NewCompressorGzip() Compressor {
	return &CompressorGzip{}
}

func (c *CompressorGzip) GetFileExtension() string {
	return ".gz"
}

func (c *CompressorGzip) NewReader(r io.Reader) (io.ReadCloser, error) {
	return gzip.NewReader(r)
}

func (c *CompressorGzip) NewWriter(w io.Writer) (io.WriteCloser, error) {
	return gzip.NewWriterLevel(w, gzip.BestCompression)
}

func init() {
	RegisterCompressor("gzip", &CompressorGzip{})
}
